
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "cartograph.h"
#include "utils.h"
#include "tokendb.h"
#include "parse.h"
#include "map.h"

struct roomrec graph[MAX_ROOMS];
char map[MAX_HEIGHT + 2][MAX_WIDTH + 3];
int numlines = 0;
int numrooms = 0;

int oppositedir[4] =
{BELOW, LEFT, ABOVE, RIGHT};
int dirxmod[10] =
{0, 1, 0, -1, 0, 0, 1, -1, 1, -1};
int dirymod[10] =
{-1, 0, 1, 0, 0, 0, -1, -1, 1, 1};
char dirstr[] = "NESWUDewbm";


void check_special(char *up, char *down, int *x, int *y, int dir)
{
  char upslide = 0, downslide = 0;
  long vector;

  vector = find_path_bitvector(map[*y][*x], oppositedir[dir]);
  if (vector & PATH_SKIP) {
    *x += dirxmod[dir];
    *y += dirymod[dir];
  }
  if (vector & PATH_UP)
    upslide = 1;
  else if (vector & PATH_DOWN)
    downslide = 1;

  if (upslide) {
    if (*down)
      *down = 0;
    else
      *up = 1;
  } else if (downslide) {
    if (*up)
      *up = 0;
    else
      *down = 1;
  }
}


int ok_location(int x, int y)
{
  if ((y > 0) && (x > 0) && (y < numlines) && (x < strlen(map[y])))
    return (1);
  return (0);
}


int follow_path(int *xl, int *yl, int *door, int dir)
{
  char up = 0, down = 0;
  char done = 0;
  int move, x, y, temp;

  x = *xl + dirxmod[dir];
  y = *yl + dirymod[dir];
  move = dir;
  *xl = *yl = -1;

  while (!done) {
    if (!ok_location(x, y))
      return (dir);
    if (get_token_room(map[y][x])) {
      *xl = x;
      *yl = y;
      done = 1;
    } else if ((temp = get_token_path(map[y][x]))) {
      if ((move = find_new_direction(temp, oppositedir[move])) == NOWHERE)
	done = 1;
      else {
	check_special(&up, &down, &x, &y, move);
	x += dirxmod[move];
	y += dirymod[move];
      }
    } else if ((temp = get_token_door(map[y][x]))) {
      if (!*door)
	*door = temp;
      x += dirxmod[move];
      y += dirymod[move];
    } else
      done = 1;
  }

  if (up)
    return (UP);
  else if (down)
    return (DOWN);
  else
    return (dir);
}


void ask_town_size(int index)
{
  int town_size = 5;
  char buf[80];

  if (graph[index].sector == SECT_CITY) {
    fprintf(stderr, "What size is the %s? ", strip_color(graph[index].name));
    fgets(buf, 80, stdin);
    if (isdigit(*buf) && atoi(buf) > 0 && atoi(buf) < 51)
      town_size = atoi(buf);
    else {
      fprintf(stderr, "Size must be 1-50!\n");
      ask_town_size(index);
    }
  }
  graph[index].town_size = town_size;
}


void init_map(char *mapfile)
{
  int index, curline, linenum;
  FILE *fp;
  char buf[MAX_WIDTH + 1];
  char *comment;

  if ((fp = fopen(mapfile, "rt")) == NULL) {
    sprintf(buf, "Input file \"%s\" not found!\n", mapfile);
    error(buf);
    exit(1);
  }
  curline = linenum = 0;
  while ((fgets(buf, MAX_WIDTH, fp) != NULL) && (curline <= MAX_HEIGHT)) {
    linenum++;
    comment = buf;
    while (isspace(*comment))
      comment++;
    if (*comment == '#')
      analyze_line(buf, linenum, mapfile);
    else {
      /* BugFix v2.01 - 05/05/94 */
      if ((comment = strchr(buf, '\n')) != NULL)
	*comment = 0;
      if ((comment = strchr(buf, ';')) != NULL)
	*comment = 0;
      sprintf(map[++curline], " %s ", buf);
    }
  }
  fclose(fp);

  if (curline > MAX_HEIGHT) {
    error("Input file too long - Raise MAX_HEIGHT constant in parse.c\n");
    exit(1);
  }
  for (index = 0; index < MAX_WIDTH; index++) {
    map[0][index] = ' ';
    map[curline + 1][index] = ' ';
  }
  numlines = curline + 1;
}


void identify_rooms(long def_flags)
{
  int index, index2, index3, roomtype, errorcnt = 0;
  char s[256];

  for (index = 1; index < numlines; index++)
    for (index2 = 1; index2 < strlen(map[index]); index2++)
      if ((roomtype = get_token_room(map[index][index2]))) {
	/* Sanity Check v3.01 - 11/14/94 */
	if (numrooms >= MAX_ROOMS) {
	  errorcnt++;
	  continue;
	}
	graph[numrooms].xcoord = index2;
	graph[numrooms].ycoord = index;
	get_room_data(map[index][index2], &graph[numrooms]);
	graph[numrooms].room_flags |= def_flags;
	for (index3 = NORTH; index3 <= SW; index3++) {
	  graph[numrooms].dir[index3].toroom = NOWHERE;
	  graph[numrooms].dir[index3].doortype = 0;
	  graph[numrooms].dir[index3].key = -1;
	  graph[numrooms].dir[index3].name = 0;
	}
	numrooms++;
      }
  /* Sanity Check v3.01 - 11/14/94 */
  if (errorcnt) {
    sprintf(s, "Too many rooms - Map truncated.\n       "
       "Raise MAX_ROOMS in cartograph.h to %d.\n", errorcnt + MAX_ROOMS);
    error(s);
  }
}


void connect_rooms()
{
  int curroom, dir, finaldir, dest, x, y, doornum;
  struct doorrec *door;

  for (curroom = 0; curroom < numrooms; curroom++)
    for (dir = NORTH; dir <= SW; dir++) {
      x = graph[curroom].xcoord;
      y = graph[curroom].ycoord;
      doornum = 0;
      finaldir = follow_path(&x, &y, &doornum, dir);
      for (dest = 0; dest < numrooms; dest++) {
	if ((graph[dest].xcoord == x) && (graph[dest].ycoord == y))
	  graph[curroom].dir[finaldir].toroom = dest;
	if (doornum) {
	  door = get_door(doornum);
	  graph[curroom].dir[finaldir].name = str_dup(door->name);
	  graph[curroom].dir[finaldir].doortype = door->doortype;
	  graph[curroom].dir[finaldir].key = door->key;
	}
      }
    }
}


void parse_map(long def_flags)
{
  identify_rooms(def_flags);
  connect_rooms();
}


void save_map(char *mapfile, int offset)
{
  FILE *fp, *fp2;
  int index, dir, counter, counter2, sector;
  char worldfile[256], buf[256], zonefile[256];
  char mapname[5];
  int mapnum;

  mapname[0] = mapfile[0];
  mapname[1] = mapfile[1];
  mapname[2] = mapfile[2];
  mapname[3] = '\0';
  mapnum = atoi(mapname);
  /* made files wrong, adjust mapname */
  mapnum += 906;
  offset = mapnum * 100;

  sprintf(zonefile, "%d.zon", mapnum);
  if ((fp2 = fopen(zonefile, "wt")) == NULL) {
    sprintf(buf, "Can not write to zone file \"%s\"\n", zonefile);
    error(buf);
    exit(1);
  }
  fprintf(fp2, "#%d\n", mapnum);
  fprintf(fp2, "Mapgrid~\n");
  fprintf(fp2, "%d 120 1\n", offset + 99);
  fprintf(fp2, "S\n$\n");
  fclose(fp2);

  sprintf(worldfile, "%d.wld", mapnum);
  if ((fp = fopen(worldfile, "wt")) == NULL) {
    sprintf(buf, "Can not write to world file \"%s\"\n", worldfile);
    error(buf);
    exit(1);
  }
  /* Added optional room name v3.10 - 04/26/95 */
  for (index = 0; index < numrooms; index++) {
    ask_town_size(index);
    fprintf(fp, "#%d\n", index + offset);
    if (graph[index].name)
      fprintf(fp, "%s~\n", graph[index].name);
    else
      fprintf(fp, "Room #%d~\n", index + offset);
    if (graph[index].sector != SECT_CITY)
      fprintf(fp, "\n");
    else
      fprintf(fp, "&+BPopulation: %d&n\n", ((graph[index].town_size * 1000) + number(-500, 500)));
    fprintf(fp, "~\n");
     /* fix a sector typo */
    if (graph[index].sector == SECT_DESERT)
      sector = (graph[index].sector + 1);
    else
      sector = graph[index].sector;
    fprintf(fp, "%d %ld %d\n", offset / 100,
	    graph[index].room_flags, sector);
    fprintf(fp, "0 0 0 0\n");
    fprintf(fp, "4 50 50 -1 0\n");
    for (dir = 0; dir < 10; dir++) {
      if (dir == UP || dir == DOWN)
	continue;
      if (graph[index].dir[dir].toroom > NOWHERE)
	graph[index].dir[dir].toroom += offset;
      if (graph[index].dir[dir].toroom > NOWHERE) {
	fprintf(fp, "D%d\n~\n", dir);
	if (graph[index].dir[dir].name)
	  fprintf(fp, "%s", graph[index].dir[dir].name);
	fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		graph[index].dir[dir].toroom);
      } else {
	if (dir == NORTH) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset - 2010);	/* unverified */

	} else if (dir == SOUTH) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset + 2010);
	} else if (dir == WEST) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset - 91);		/* verified */
	} else if (dir == EAST) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset + 91);
	} else if (dir == NE) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset - 101);
	} else if (dir == SE) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset + 101);
	} else if (dir == SW) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset - 81);
	} else if (dir == NW) {
	  fprintf(fp, "D%d\n~\n", dir);
	  fprintf(fp, "~\n0 -1 %d 0 0 1\n",
		  index + offset - 2101);
        }
      }
    }
/*** create/draw maps ***/
    fprintf(fp, "M\n");
    for (counter = 0; counter < 50; counter++) {
      for (counter2 = 0; counter2 < 50; counter2++) {
	fprintf(fp, " %d", map_symbol(index, offset, counter, counter2));
      }
      fprintf(fp, "\n");
    }
/*** end of map drawings ***/
    fprintf(fp, "S\n");
  }
  fprintf(fp, "$\n");
  printf("Completed.....\n");
  fclose(fp);
}

/* pass in the coord, check nearby symbols and sector, return new symbol */
int map_symbol(int index, int offset, int x, int y)
{
  int quadrant = -1;		/* this decides which dir we care to use to altar */
  int rock, hill, mount, tree, grass, lake, river, brush, cactus;
  int symbol = MAP_BLANK;
  int choices = 0, choice[10], sector_mod, size;
  int n, t, t2;
  int returning_water = FALSE;
  char buf[80];

  /* clear choices */
  for (n = 0; n < 10; n++)
    choice[n] = MAP_BLANK;

  /* 2% chance sector changes to one of these */
  tree = grass = 4;
  cactus = 0;
  hill = 0;
  rock = 0;
  brush = 0;
  mount = river = lake = 0;

  /* lakes, rivers are a special case. deal with them now. */
  if (graph[index].sector == SECT_WATER_SWIM) {
    if (x <= 20) {
      if (y <= 20)
	quadrant = NW;
      else if (y >= 30)
	quadrant = NE;
      else
	quadrant = NORTH;
    } else if (x >= 30) {
      if (y <= 20)
	quadrant = SW;
      else if (y >= 30)
	quadrant = SE;
      else
	quadrant = SOUTH;
    } else if (y <= 20)
      quadrant = WEST;
    else if (y >= 30)
      quadrant = EAST;

    /* first round, lake in center then loop cardinals. */
    if (quadrant == -1)
      return MAP_LAKE;		/* this one is a must, so we'll return it now */
    if (quadrant == SOUTH && graph[index].dir[SOUTH].toroom > NOWHERE &&
	graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == NORTH && graph[index].dir[NORTH].toroom > NOWHERE &&
	graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == WEST && graph[index].dir[WEST].toroom > NOWHERE &&
	graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == EAST && graph[index].dir[EAST].toroom > NOWHERE &&
	graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;

    /* second round, redo quads and loop diagonals */
    t = y - x;
    t2 = y + x;
    quadrant = -1;		/* reset to center */
    if (t < -10) {
      if (t2 < 40)
	quadrant = WEST;
      else if (t2 < 60)
	quadrant = SW;
      else
	quadrant = SOUTH;
    } else if (t < 10) {
      if (t2 < 40)
	quadrant = NW;
      else if (t2 < 60)
	quadrant = -1;
      else
	quadrant = SE;
    } else {
      if (t2 < 40)
	quadrant = NORTH;
      else if (t2 < 60)
	quadrant = NE;
      else
	quadrant = EAST;
    }
    if (quadrant == NW && graph[index].dir[NW].toroom > NOWHERE &&
    graph[graph[index].dir[NW].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == SE && graph[index].dir[SE].toroom > NOWHERE &&
    graph[graph[index].dir[SE].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == NE && graph[index].dir[NE].toroom > NOWHERE &&
    graph[graph[index].dir[NE].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == SW && graph[index].dir[SW].toroom > NOWHERE &&
    graph[graph[index].dir[SW].toroom - offset].sector == SECT_WATER_SWIM)
      returning_water = TRUE;
    if (quadrant == -1)
      returning_water = TRUE;
  }
  /* back to normal checks */
  /* This one should be for non water rooms, that need a small tip of
     water for a diagonal connection */
  /* also used to adjust terrain near the edge of a map */
  if (y <= (-x + 10))
    quadrant = NW;
  else if (y >= (x + 40))
    quadrant = NE;
  else if (y >= (-x + 60))
    quadrant = SE;
  else if (y <= (x - 40))
    quadrant = SW;
  else if (x <= 10)
    quadrant = NORTH;
  else if (x >= 40)
    quadrant = SOUTH;
  else if (y <= 10)
    quadrant = WEST;
  else if (y >= 40)
    quadrant = EAST;

  /* initialize base symbol */
  switch (graph[index].sector) {
  case SECT_FOREST:
    choice[0] = MAP_TREE;
    grass += 80;
    brush += 100;
    break;
  case SECT_SWAMP:
    choice[0] = MAP_SWAMP;
    break;
  case SECT_FIELD:
    choice[0] = MAP_FIELD2;
    break;
  case SECT_CITY:
    if (graph[index].dir[NORTH].toroom > NOWHERE) {
      if (graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_FOREST)
	choice[0] = MAP_TREE;
      else if (graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_HILLS)
	choice[0] = MAP_HILL;
      else if (graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_MOUNTAIN)
	choice[0] = MAP_MOUNTAIN;
      else if (graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_DESERT)
	choice[0] = MAP_DESERT;
      else
	choice[0] = MAP_FIELD2;
    } else
      choice[0] = MAP_FIELD2;
    break;
  case SECT_HILLS:
    choice[0] = MAP_HILL;
    break;
  case SECT_MOUNTAIN:
    choice[0] = MAP_MOUNTAIN;
    break;
  case SECT_DESERT:
    choice[0] = MAP_DESERT;
    break;
  case SECT_WATER_SWIM:
    choice[0] = MAP_FIELD;	/* already returned lake where we should */
    break;
  case SECT_WATER_NOSWIM:
    choice[0] = MAP_LAKE;
    break;
  }

  /* its near a side, adjust chance of changing symbol */
  if (quadrant > -1 && graph[index].dir[quadrant].toroom > NOWHERE) {
    sector_mod = graph[graph[index].dir[quadrant].toroom - offset].sector;
    switch (sector_mod) {
    case SECT_FOREST:
      tree += 10;
      brush += 15;
      break;
    case SECT_FIELD:
      grass += 15;
      break;
    case SECT_DESERT:
      rock += 5;
      cactus += 15;
      break;
    case SECT_HILLS:
      mount += 10;
      rock += 5;
      break;
    case SECT_MOUNTAIN:
      hill += 10;
      rock += 5;
      break;
    case SECT_WATER_SWIM:
    case SECT_WATER_NOSWIM:
      break;
    }
  }
  if (graph[index].sector == SECT_WATER_SWIM || graph[index].sector == SECT_CITY) {
    /* terrain will be as other room is */
    switch (sector_mod) {
    case SECT_FOREST:
      choice[0] = MAP_TREE;
      grass += 80;
      brush += 100;
      break;
    case SECT_FIELD:
      choice[0] = MAP_FIELD2;
      break;
    case SECT_HILLS:
      choice[0] = MAP_HILL;
      break;
    case SECT_MOUNTAIN:
      choice[0] = MAP_MOUNTAIN;
      break;
    case SECT_DESERT:
      choice[0] = MAP_DESERT;
      break;
    case SECT_WATER_SWIM:
      choice[0] = MAP_FIELD2;
      break;
    }
  }
  /* this is the triangle tip producer */
  if (quadrant == SE && graph[index].dir[SOUTH].toroom > NOWHERE && graph[index].dir[EAST].toroom > NOWHERE &&
      graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM &&
  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM)
    returning_water = TRUE;
  if (quadrant == NE && graph[index].dir[NORTH].toroom > NOWHERE && graph[index].dir[EAST].toroom > NOWHERE &&
      graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM &&
  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM)
    returning_water = TRUE;
  if (quadrant == SW && graph[index].dir[SOUTH].toroom > NOWHERE && graph[index].dir[WEST].toroom > NOWHERE &&
      graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM &&
  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM)
    returning_water = TRUE;
  if (quadrant == NW && graph[index].dir[NORTH].toroom > NOWHERE && graph[index].dir[WEST].toroom > NOWHERE &&
      graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM &&
  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM)
    returning_water = TRUE;

  if (number(0, 100) < rock) {
    choices++;
    choice[choices] = MAP_ROCK;
  }
  if (number(0, 100) < hill) {
    choices++;
    choice[choices] = MAP_HILL;
  }
  if (number(0, 100) < mount) {
    choices++;
    choice[choices] = MAP_MOUNTAIN;
  }
  if (number(0, 100) < tree) {
    choices++;
    choice[choices] = MAP_TREE;
  }
  if (number(0, 100) < cactus) {
    choices++;
    choice[choices] = MAP_CACTUS;
  }
  if (number(0, 100) < grass) {
    choices++;
    choice[choices] = MAP_FIELD2;
  }
  if (number(0, 100) < brush) {
    choices++;
    choice[choices] = MAP_BRUSH;
  }
  /* lets try to fix the goofyness right here */
  if (returning_water) {
    if (x <= 20) {
      if (y < 20)
	quadrant = NW;
      else if (y > 30)
	quadrant = NE;
      else
	quadrant = NORTH;
    } else if (x >= 30) {
      if (y < 20)
	quadrant = SW;
      else if (y > 30)
	quadrant = SE;
      else
	quadrant = SOUTH;
    } else if (y <= 20)
      quadrant = WEST;
    else if (y >= 30)
      quadrant = EAST;
    /* The if from hell */
    if (
	 (graph[index].dir[SE].toroom > NOWHERE &&
    graph[graph[index].dir[SE].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == NORTH) ||
	 (graph[index].dir[NE].toroom > NOWHERE &&
    graph[graph[index].dir[NE].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == WEST) ||
	 (graph[index].dir[NW].toroom > NOWHERE &&
    graph[graph[index].dir[NW].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == SOUTH) ||
	 (graph[index].dir[SW].toroom > NOWHERE &&
    graph[graph[index].dir[SW].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == EAST) ||
	 (graph[index].dir[SE].toroom > NOWHERE &&
    graph[graph[index].dir[SE].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == WEST) ||
	 (graph[index].dir[NE].toroom > NOWHERE &&
    graph[graph[index].dir[NE].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == SOUTH) ||
	 (graph[index].dir[NW].toroom > NOWHERE &&
    graph[graph[index].dir[NW].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == EAST) ||
	 (graph[index].dir[SW].toroom > NOWHERE &&
    graph[graph[index].dir[SW].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM
	  && quadrant == NORTH) ||
	 (graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_SWIM
	  && (quadrant == NORTH || quadrant == SOUTH)) ||
	 (graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_SWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_SWIM
	  && (quadrant == WEST || quadrant == EAST))
	)
      returning_water = FALSE;
  }
  /* end of goofyness repair */


  /* Make diagonal edges of ocean as needed */
  quadrant = -1;
  if (y <= (-x + 25))
    quadrant = NW;
  else if (y >= (x + 25))
    quadrant = NE;
  else if (y >= (-x + 75))
    quadrant = SE;
  else if (y <= (x - 25))
    quadrant = SW;
  if (graph[index].sector == SECT_WATER_NOSWIM) {	/* ocean */
    if (
	 (graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector != SECT_WATER_NOSWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector != SECT_WATER_NOSWIM
	  && quadrant == NW) ||
	 (graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector != SECT_WATER_NOSWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector != SECT_WATER_NOSWIM
	  && quadrant == SW) ||
	 (graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector != SECT_WATER_NOSWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector != SECT_WATER_NOSWIM
	  && quadrant == NE) ||
	 (graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector != SECT_WATER_NOSWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector != SECT_WATER_NOSWIM
	  && quadrant == SE)
	) {
      switch (graph[graph[index].dir[quadrant].toroom - offset].sector) {
      case SECT_FOREST:
	choice[0] = MAP_TREE;
	grass += 80;
	brush += 100;
	break;
      case SECT_FIELD:
	choice[0] = MAP_FIELD2;
	break;
      case SECT_DESERT:
        choice[0] = MAP_DESERT;
        break;
      case SECT_HILLS:
	choice[0] = MAP_HILL;
	break;
      case SECT_MOUNTAIN:
	choice[0] = MAP_MOUNTAIN;
	break;
      case SECT_WATER_SWIM:
	choice[0] = MAP_LAKE;
	break;
      }
    } else
      choices = 0;
  } else {
    if (
	 (graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_NOSWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_NOSWIM
	  && quadrant == NW) ||
	 (graph[index].dir[WEST].toroom > NOWHERE &&
	  graph[graph[index].dir[WEST].toroom - offset].sector == SECT_WATER_NOSWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_NOSWIM
	  && quadrant == SW) ||
	 (graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_NOSWIM
	  && graph[index].dir[NORTH].toroom > NOWHERE &&
	  graph[graph[index].dir[NORTH].toroom - offset].sector == SECT_WATER_NOSWIM
	  && quadrant == NE) ||
	 (graph[index].dir[EAST].toroom > NOWHERE &&
	  graph[graph[index].dir[EAST].toroom - offset].sector == SECT_WATER_NOSWIM
	  && graph[index].dir[SOUTH].toroom > NOWHERE &&
	  graph[graph[index].dir[SOUTH].toroom - offset].sector == SECT_WATER_NOSWIM
	  && quadrant == SE)
	)
      returning_water = TRUE;
  }
  /* town check, draw walls based on size */
  if (graph[index].sector == SECT_CITY) {
    size = graph[index].town_size;
    if ((x < ((50 - (size)) / 2) || x > ((size) + ((50 - (size)) / 2)))
	     && (y == 25))
      symbol = MAP_ROAD;
    else if ((y < ((50 - (size)) / 2) || y > ((size) + ((50 - (size)) / 2)))
	     && (x == 25))
      symbol = MAP_ROAD;
    else if (x < ((50 - (size)) / 2) || x > ((size) + ((50 - (size)) / 2)))
      symbol = choice[number(0, choices)];
    else if (y < ((50 - (size)) / 2) || y > ((size) + ((50 - (size)) / 2)))
      symbol = choice[number(0, choices)];
    else if ((x == ((50 - (size)) / 2) || x == ((size) + ((50 - (size)) / 2)))
	     && (y == 25))
      symbol = MAP_OPEN_DOOR;
    else if ((y == ((50 - (size)) / 2) || y == ((size) + ((50 - (size)) / 2)))
	     && (x == 25))
      symbol = MAP_OPEN_DOOR;
    else if ((x == ((50 - (size)) / 2) || x == ((size) + ((50 - (size)) / 2)))
    && (y >= ((50 - (size)) / 2) || y <= ((size) + ((50 - (size)) / 2))))
      symbol = MAP_WALL;
    else if ((y == ((50 - (size)) / 2) || y == ((size) + ((50 - (size)) / 2)))
    && (x >= ((50 - (size)) / 2) || x <= ((size) + ((50 - (size)) / 2))))
      symbol = MAP_WALL;
    else
      symbol = MAP_NORMAL_FLOOR;
    /* This repairs south and east walls on size 50 cities */
    if (symbol == MAP_NORMAL_FLOOR && (y == 50 || x == 50))
      if (x == 50 && y == 25)
        symbol = MAP_OPEN_DOOR;
      else if (y == 50 && x == 25)
        symbol = MAP_OPEN_DOOR;
      else
        symbol = MAP_WALL;
  } else
    symbol = choice[number(0, choices)];
  if (returning_water)
    return MAP_LAKE;
  return symbol;
}
